extern crate core;

use std::fmt::format;
use std::process::Command;
use std::thread::sleep;
use std::time::Duration;
use serde::Deserialize;
use figment::{Figment, providers::{Format, Toml}};

#[derive(Deserialize)]
struct Config {
    rsync_password: String,
    rsync_url: String,
    hydrus_db_location: String,
    hydrus_exec_location: String
}

impl Config {
    pub fn make_rsync_url(&self, endpoint: &str) -> String {
        return format!("{}/{}", self.rsync_url, endpoint);
    }
}

fn make_temp_location(path: &str) -> String {
    let en = std::env::temp_dir();
    let dir = en.to_str().unwrap();
    return format!("{}/{}", dir, path);
}

fn launch_hydrus(config: &Config) {
    println!("Launching hydrus...");

    let mut hydrus = Command::new(&config.hydrus_exec_location);
    let hydrus = hydrus.args(["-d", &config.hydrus_db_location]);

    if cfg!(target_os = "linux") {
        hydrus
            .env("QT_QPA_PLATFORM", "xcb")
            .status()
            .expect("Failed to run hydrus");
    } else {
        hydrus
            .status()
            .expect("Failed to run hydrus");
    }
}

fn is_locked(config: &Config) -> bool {
    println!("Receiving lock... ");

    let output = Command::new("rsync")
        .args([config.make_rsync_url("lock"), make_temp_location("lock")])
        .env("RSYNC_PASSWORD", &config.rsync_password)
        .status()
        .expect("Failed to sync lockfile.");

    // TODO: check this better?
    if !output.success() {
        panic!("Failed to sync lockfile!");
    }

    let mut str = std::fs::read_to_string(make_temp_location("lock")).unwrap();
    str = str.trim().parse().unwrap();
    println!("{}", str);

    str == "1"
}

fn set_lock(config: &Config, locked: bool) {
    if locked {
        std::fs::write(make_temp_location("lock"), "1").unwrap();
    } else {
        std::fs::write(make_temp_location("lock"), "0").unwrap();
    }

    let output = Command::new("rsync")
        .args([make_temp_location("lock"), config.make_rsync_url("lock")])
        .env("RSYNC_PASSWORD", &config.rsync_password)
        .status()
        .expect("Failed to sync lockfile.");
}

fn get_last_hostname(config: &Config) -> Option<String> {
    println!("Receiving last hostname... ");

    let output = Command::new("rsync")
        .args([config.make_rsync_url("lasthost"), make_temp_location("lasthost")])
        .env("RSYNC_PASSWORD", &config.rsync_password)
        .status()
        .expect("Failed to sync hostname.");

    // TODO: check this better?
    if !output.success() {
        panic!("Failed to sync hostname!");
    }

    std::fs::read_to_string(make_temp_location("lasthost")).ok()
}

fn set_last_hostname(config: &Config, hostname: String) {
    std::fs::write(make_temp_location("lasthost"), hostname).unwrap();

    let output = Command::new("rsync")
        .args([make_temp_location("lasthost"), config.make_rsync_url("lasthost")])
        .env("RSYNC_PASSWORD", &config.rsync_password)
        .status()
        .expect("Failed to sync hostname.");
}

fn get_current_hostname() -> Option<String> {
    println!("Checking current hostname... ");
    hostname::get().ok()?.into_string().ok()
}

fn sync_files(config: &Config, from: &str, to: &str) {
    // send first batch of files based on size alone
    // works for media, since that doesn't change
    // TODO: exclude database files entirely at this stage
    Command::new("rsync")
        .args(
            ["-rz",
            "--size-only",
            "-hh",
            "--partial",
            "--info=stats1,progress2",
            "--modify-window=1",
            "--exclude=@eaDir",
            from,
            to])
        .env("RSYNC_PASSWORD", &config.rsync_password)
        .status()
        .expect("Failed to sync media files.");

    // now this only includes database files
    // this explicitly ignores size, as two databases can be the same "size" but may have different
    // contents? i think that's why i wrote this lol
    Command::new("rsync")
        .args(
            ["-rzcv",
                "--include='*\\/'",
                "--include='*.db'",
                "--exclude='*'",
                "--exclude=@eaDir",
                from,
                to])
        .env("RSYNC_PASSWORD", &config.rsync_password)
        .status()
        .expect("Failed to sync media files.");

    println!("Synchronization complete!");
}

fn main() {
    println!("Starting hydrus-sync...");

    let config: Config = Figment::new()
        .merge(Toml::file("hydrus-sync.toml"))
        .extract().unwrap();

    let mut last_hostname = get_last_hostname(&config)
        .expect("Could not get last hostname.");
    last_hostname = last_hostname.trim().parse().unwrap();

    let mut current_hostname = get_current_hostname()
        .expect("Could not get current hostname.");
    current_hostname = current_hostname.trim().parse().unwrap();

    if is_locked(&config) {
        println!("Database is currently in use!");
        return;
    }

    println!("Current hostname: {}", current_hostname);
    println!("Last hostname: {}", last_hostname);

    if current_hostname == last_hostname {
        println!("Sync is not needed.");
    } else {
        println!("Sync is required.");

        sync_files(&config, &config.rsync_url, &config.hydrus_db_location);
    }

    set_lock(&config, true);

    launch_hydrus(&config);

    println!("Hydrus has closed. Beginning sync...");

    sync_files(&config, &config.hydrus_db_location, &config.rsync_url);

    set_last_hostname(&config, current_hostname);
    set_lock(&config, false);
}
